"""
Test ConfigWatcher signal receivers (main code).
"""

import json
from contextlib import ExitStack
from unittest.mock import call, patch

import ddt
from django.test import TestCase, override_settings

from edx_arch_experiments.config_watcher.signals import receivers


@ddt.ddt
class TestConfigWatcherReceivers(TestCase):

    @ddt.unpack
    @ddt.data(
        (
            None, None, None,
        ),
        (
            'https://localhost', None, "test message",
        ),
        (
            'https://localhost', 'my-ida', "[my-ida] test message",
        ),
    )
    def test_send_to_slack(self, slack_url, service_name, expected_message):
        """Check that message prefixing is performed as expected."""
        # This can be made cleaner in Python 3.10
        with ExitStack() as stack:
            patches = [
                patch('urllib.request.Request'),
                patch('urllib.request.urlopen'),
                patch.object(receivers, 'CONFIG_WATCHER_SLACK_WEBHOOK_URL', slack_url),
                patch.object(receivers, 'CONFIG_WATCHER_SERVICE_NAME', service_name),
            ]
            (mock_req, _, _, _) = [stack.enter_context(cm) for cm in patches]
            receivers._send_to_slack("test message")

        if expected_message is None:
            mock_req.assert_not_called()
        else:
            mock_req.assert_called_once()
            (call_args, call_kwargs) = mock_req.call_args_list[0]
            assert json.loads(call_kwargs['data'])['text'] == expected_message



"""
Utilities for monitoring code_owner_2
"""
import logging
import re

from django.conf import settings
from edx_django_utils.monitoring import set_custom_attribute

log = logging.getLogger(__name__)


def get_code_owner_from_module(module):
    """
    Attempts lookup of code_owner based on a code module,
    finding the most specific match. If no match, returns None.

    For example, if the module were 'openedx.features.discounts.views',
    this lookup would match on 'openedx.features.discounts' before
    'openedx.features', because the former is more specific.

    See how to:
    https://github.com/openedx/edx-django-utils/blob/master/edx_django_utils/monitoring/docs/how_tos/add_code_owner_custom_attribute_to_an_ida.rst

    """
    if not module:
        return None

    code_owner_mappings = get_code_owner_mappings()
    if not code_owner_mappings:
        return None

    module_parts = module.split('.')
    # To make the most specific match, start with the max number of parts
    for number_of_parts in range(len(module_parts), 0, -1):
        partial_path = '.'.join(module_parts[0:number_of_parts])
        if partial_path in code_owner_mappings:
            code_owner = code_owner_mappings[partial_path]
            return code_owner
    return None


def is_code_owner_mappings_configured():
    """
    Returns True if code owner mappings were configured, and False otherwise.
    """
    return isinstance(get_code_owner_mappings(), dict)


# cached lookup table for code owner given a module path.
# do not access this directly, but instead use get_code_owner_mappings.
_PATH_TO_CODE_OWNER_MAPPINGS = None


def get_code_owner_mappings():
    """
    Returns the contents of the CODE_OWNER_MAPPINGS Django Setting, processed
    for efficient lookup by path.

    Returns:
         (dict): dict mapping modules to code owners, or None if there are no
            configured mappings, or an empty dict if there is an error processing
            the setting.

    Example return value::

        {
            'xblock_django': 'team-red',
            'openedx.core.djangoapps.xblock': 'team-red',
            'badges': 'team-blue',
        }

    """
    global _PATH_TO_CODE_OWNER_MAPPINGS

    # Return cached processed mappings if already processed
    if _PATH_TO_CODE_OWNER_MAPPINGS is not None:
        return _PATH_TO_CODE_OWNER_MAPPINGS

    # Uses temporary variable to build mappings to avoid multi-threading issue with a partially
    # processed map.  Worst case, it is processed more than once at start-up.
    path_to_code_owner_mapping = {}

    # .. setting_name: CODE_OWNER_MAPPINGS
    # .. setting_default: None
    # .. setting_description: Used for monitoring and reporting of ownership. Use a
    #      dict with keys of code owner name and value as a list of dotted path
    #      module names owned by the code owner.
    code_owner_mappings = getattr(settings, 'CODE_OWNER_MAPPINGS', None)
    if code_owner_mappings is None:
        return None

    try:
        for code_owner in code_owner_mappings:
            path_list = code_owner_mappings[code_owner]
            for path in path_list:
                path_to_code_owner_mapping[path] = code_owner
                optional_module_prefix_match = _OPTIONAL_MODULE_PREFIX_PATTERN.match(path)
                # if path has an optional prefix, also add the module name without the prefix
                if optional_module_prefix_match:
                    path_without_prefix = path[optional_module_prefix_match.end():]
                    path_to_code_owner_mapping[path_without_prefix] = code_owner
    except TypeError as e:
        log.exception(
            'Error processing CODE_OWNER_MAPPINGS. {}'.format(e)  # pylint: disable=logging-format-interpolation
        )
        raise e

    _PATH_TO_CODE_OWNER_MAPPINGS = path_to_code_owner_mapping
    return _PATH_TO_CODE_OWNER_MAPPINGS


def set_code_owner_attribute_from_module(module):
    """
    Updates the code_owner_2 and code_owner_2_module custom attributes.

    Celery tasks or other non-web functions do not use middleware, so we need
        an alternative way to set the code_owner_2 custom attribute.

    Note: These settings will be overridden by the CodeOwnerMonitoringMiddleware.
        This method can't be used to override web functions at this time.

    Usage::

        set_code_owner_2_attribute_from_module(__name__)

    """
    set_custom_attribute('code_owner_2_module', module)
    code_owner = get_code_owner_from_module(module)

    if code_owner:
        set_code_owner_custom_attributes(code_owner)


def set_code_owner_custom_attributes(code_owner):
    """
    Sets custom metrics for code_owner_2, code_owner_2_theme, and code_owner_2_squad
    """
    if not code_owner:  # pragma: no cover
        return
    set_custom_attribute('code_owner_2', code_owner)
    theme = _get_theme_from_code_owner(code_owner)
    if theme:
        set_custom_attribute('code_owner_2_theme', theme)
    squad = _get_squad_from_code_owner(code_owner)
    if squad:
        set_custom_attribute('code_owner_2_squad', squad)


def clear_cached_mappings():
    """
    Clears the cached code owner mappings. Useful for testing.
    """
    global _PATH_TO_CODE_OWNER_MAPPINGS
    _PATH_TO_CODE_OWNER_MAPPINGS = None
    global _CODE_OWNER_TO_THEME_AND_SQUAD_MAPPINGS
    _CODE_OWNER_TO_THEME_AND_SQUAD_MAPPINGS = None


# TODO: Retire this once edx-platform import_shims is no longer used.
#   Note: This should be ready for removal because import_shims has been removed.
#   See https://github.com/openedx/edx-platform/tree/854502b560bda74ef898501bb2a95ce238cf794c/import_shims
_OPTIONAL_MODULE_PREFIX_PATTERN = re.compile(r'^(lms|common|openedx\.core)\.djangoapps\.')


# Cached lookup table for code owner theme and squad given a code owner.
# - Although code owner is "theme-squad", a hyphen may also be in the theme or squad name, so this ensures we get both
#   correctly from config.
# Do not access this directly, but instead use get_code_owner_theme_squad_mappings.
_CODE_OWNER_TO_THEME_AND_SQUAD_MAPPINGS = None


def get_code_owner_theme_squad_mappings():
    """
    Returns the contents of the CODE_OWNER_THEMES Django Setting, processed
    for efficient lookup by path.

    Returns:
         (dict): dict mapping code owners to a dict containing the squad and theme, or
            an empty dict if there are no configured mappings.

    Example return value::

        {
            'theme-x-team-red': {
                'theme': 'theme-x',
                'squad': 'team-red',
            },
            'theme-x-team-blue': {
                'theme': 'theme-x',
                'squad': 'team-blue',
            },
        }

    """
    global _CODE_OWNER_TO_THEME_AND_SQUAD_MAPPINGS

    # Return cached processed mappings if already processed
    if _CODE_OWNER_TO_THEME_AND_SQUAD_MAPPINGS is not None:
        return _CODE_OWNER_TO_THEME_AND_SQUAD_MAPPINGS

    # Uses temporary variable to build mappings to avoid multi-threading issue with a partially
    # processed map.  Worst case, it is processed more than once at start-up.
    code_owner_to_theme_and_squad_mapping = {}

    # .. setting_name: CODE_OWNER_THEMES
    # .. setting_default: None
    # .. setting_description: Used for monitoring and reporting of ownership. Use a
    #      dict with keys of code owner themes and values as a list of code owner names
    #      including theme and squad, separated with a hyphen.
    code_owner_themes = getattr(settings, 'CODE_OWNER_THEMES', {})

    try:
        for theme in code_owner_themes:
            code_owner_list = code_owner_themes[theme]
            for code_owner in code_owner_list:
                squad = code_owner.split(theme + '-', 1)[1]
                code_owner_details = {
                    'theme': theme,
                    'squad': squad,
                }
                code_owner_to_theme_and_squad_mapping[code_owner] = code_owner_details
    except TypeError as e:
        log.exception(
            'Error processing CODE_OWNER_THEMES setting. {}'.format(e)  # pylint: disable=logging-format-interpolation
        )
        raise e

    _CODE_OWNER_TO_THEME_AND_SQUAD_MAPPINGS = code_owner_to_theme_and_squad_mapping
    return _CODE_OWNER_TO_THEME_AND_SQUAD_MAPPINGS


def _get_theme_from_code_owner(code_owner):
    """
    Returns theme for a code_owner (e.g. 'theme-my-squad' => 'theme')
    """
    mappings = get_code_owner_theme_squad_mappings()
    if mappings is None:  # pragma: no cover
        return None

    if code_owner in mappings:
        return mappings[code_owner]['theme']

    return None


def _get_squad_from_code_owner(code_owner):
    """
    Returns squad for a code_owner (e.g. 'theme-my-squad' => 'my-squad')
    """
    mappings = get_code_owner_theme_squad_mappings()
    if mappings is None:  # pragma: no cover
        return None

    if code_owner in mappings:
        return mappings[code_owner]['squad']

    return None



"""
Tests for celery signal handler.
"""
from ddtrace import tracer
from django.test import TestCase

from edx_arch_experiments.datadog_monitoring.signals.handlers import init_worker_process


class TestHandlers(TestCase):
    """Tests for signal handlers."""

    def setUp(self):
        # Remove custom span processor from previous runs.
        # pylint: disable=protected-access
        tracer._span_processors = [
            sp for sp in tracer._span_processors if type(sp).__name__ != 'CeleryCodeOwnerSpanProcessor'
        ]

    def test_init_worker_process(self):
        def get_processor_list():
            # pylint: disable=protected-access
            return [type(sp).__name__ for sp in tracer._span_processors]

        assert sorted(get_processor_list()) == [
            'EndpointCallCounterProcessor', 'TopLevelSpanProcessor',
        ]

        init_worker_process(sender=None)

        assert sorted(get_processor_list()) == [
            'CeleryCodeOwnerSpanProcessor', 'EndpointCallCounterProcessor', 'TopLevelSpanProcessor',
        ]



"""
Tests for plugin app.
"""
from celery.signals import worker_process_init
from django.test import TestCase


class TestDatadogMonitoringApp(TestCase):
    """
    Tests for TestDatadogMonitoringApp.
    """

    def test_signal_has_receiver(self):
        """
        Imperfect test to ensure celery signal has the receiver.

        The receiver gets added during DatadogMonitoringApp's ready() call
        at load time. An attempt to disconnect the receiver did not allow it
        to be re-added during a call to ready, presumably because the signal
        was already imported.
        """
        # the name of the function is in the weakref __repr__
        assert 'init_worker_process' in repr(worker_process_init.receivers[0][1])



"""
Test codejail service views.
"""

import json
import textwrap
from os import path

import ddt
from django.contrib.auth import get_user_model
from django.test import TestCase, override_settings
from django.urls import reverse
from rest_framework.test import APIClient

from edx_arch_experiments.codejail_service import views


@override_settings(
    ROOT_URLCONF='edx_arch_experiments.codejail_service.urls',
    MIDDLEWARE=[
        'django.contrib.sessions.middleware.SessionMiddleware',
    ],
)
@ddt.ddt
class TestExecService(TestCase):
    """Test the v0 code exec view."""

    def setUp(self):
        super().setUp()
        user_model = get_user_model()
        self.admin_user = user_model.objects.create_user('cms_worker', is_staff=True)
        self.other_user = user_model.objects.create_user('student', is_staff=False)
        self.standard_params = {'code': 'retval = 3 + 4', 'globals_dict': {}}

    def _test_codejail_api(self, *, user=None, skip_auth=False, params=None, files=None, exp_status, exp_body):
        """
        Call the view and make assertions.

        Args:
            user: User to authenticate as when calling view, defaulting to an is_staff user
            skip_auth: If true, do not send authentication headers (incompatible with `user` argument)
            params: Payload of codejail parameters, defaulting to a simple arithmetic check
            files: Files to include in the API call, as dict of filenames to file objects
            exp_status: Assert that the response HTTP status code is this value
            exp_body: Assert that the response body JSON is this value
        """
        assert not (user and skip_auth)

        client = APIClient()
        user = user or self.admin_user
        if not skip_auth:
            client.force_authenticate(user)

        params = self.standard_params if params is None else params
        payload = json.dumps(params)
        req_body = {'payload': payload, **(files or {})}

        resp = client.post(reverse('code_exec_v0'), req_body, format='multipart')

        assert resp.status_code == exp_status
        assert json.loads(resp.content) == exp_body

    def test_success(self):
        """Regular successful call."""
        self._test_codejail_api(
            exp_status=200, exp_body={'globals_dict': {'retval': 7}},
        )

    @override_settings(CODEJAIL_SERVICE_ENABLED=False)
    def test_feature_disabled(self):
        """Service can be disabled."""
        self._test_codejail_api(
            exp_status=500, exp_body={'error': "Codejail service not enabled"},
        )

    @override_settings(ENABLE_CODEJAIL_REST_SERVICE=True)
    def test_misconfigured_as_relay(self):
        """Don't accept codejail requests if we're going to send them elsewhere."""
        self._test_codejail_api(
            exp_status=500, exp_body={'error': "Codejail service is misconfigured. (Refusing to act as relay.)"},
        )

    def test_unauthenticated(self):
        """Anonymous requests are rejected."""
        self._test_codejail_api(
            skip_auth=True,
            exp_status=403, exp_body={'detail': "Authentication credentials were not provided."},
        )

    def test_unprivileged(self):
        """Anonymous requests are rejected."""
        self._test_codejail_api(
            user=self.other_user,
            exp_status=403, exp_body={'detail': "You do not have permission to perform this action."},
        )

    def test_unsafely(self):
        """unsafely=true is rejected"""
        self._test_codejail_api(
            params=dict(**self.standard_params, unsafely=True),
            exp_status=400, exp_body={'error': "Refusing codejail execution with unsafely=true"},
        )

    @ddt.unpack
    @ddt.data(
        ({'globals_dict': {}}, 'code'),
        ({'code': 'retval = 3 + 4'}, 'globals_dict'),
        ({}, 'code'),
    )
    def test_missing_params(self, params, missing):
        """Two code and globals_dict params are required."""
        self._test_codejail_api(
            params=params,
            exp_status=400, exp_body={
                'error': f"Payload JSON did not match schema: '{missing}' is a required property",
            },
        )

    def test_extra_files(self):
        """Check that we can include a course library."""
        # "Course library" containing `course_library.triangular_number`.
        #
        # It's tempting to use zipfile to write to an io.BytesIO so
        # that the test library is in plaintext. Django's request
        # factory will indeed see that as a file to use in a multipart
        # upload, but it will see it as an empty bytestring. (read()
        # returns empty bytestring, while getvalue() returns the
        # desired data). So instead we just have a small zip file on
        # disk here.
        library_path = path.join(path.dirname(__file__), 'test_course_library.zip')

        with open(library_path, 'rb') as lib_zip:
            self._test_codejail_api(
                params={
                    'code': textwrap.dedent("""
                        from course_library import triangular_number

                        result = triangular_number(6)
                    """),
                    'globals_dict': {},
                    'python_path': ['python_lib.zip'],
                },
                files={'python_lib.zip': lib_zip},
                exp_status=200, exp_body={'globals_dict': {'result': 21}},
            )

    def test_exception(self):
        """Report exceptions from jailed code."""
        self._test_codejail_api(
            params={'code': '1/0', 'globals_dict': {}},
            exp_status=200, exp_body={'emsg': 'ZeroDivisionError: division by zero'},
        )



"""
Datadog span processor for celery span code owners.
"""
from .utils import set_code_owner_attribute_from_module


class CeleryCodeOwnerSpanProcessor:
    """
    Datadog span processor that adds celery code owner span tags.
    """

    def on_span_start(self, span):
        """
        Adds code owner span tag for celery run spans at span creation.
        """
        if getattr(span, 'name', None) == 'celery.run':
            # We can use this for celery spans, because the resource name is more predictable
            # and available from the start. For django requests, we'll instead continue to use
            # django middleware for setting code owner.
            set_code_owner_attribute_from_module(span.resource)

    def on_span_finish(self, span):
        pass

    def shutdown(self, _timeout):
        pass



"""
Handlers to listen to celery signals.
"""
import logging

from celery.signals import worker_process_init
from django.dispatch import receiver

from edx_arch_experiments.datadog_monitoring.code_owner.datadog import CeleryCodeOwnerSpanProcessor

log = logging.getLogger(__name__)


@receiver(worker_process_init)
def init_worker_process(sender, **kwargs):
    """
    Adds a Datadog span processor to each worker process.

    We have to do this from inside the worker processes because they fork from the
    parent process before the plugin app is initialized.
    """
    try:
        from ddtrace import tracer  # pylint: disable=import-outside-toplevel

        tracer._span_processors.append(CeleryCodeOwnerSpanProcessor())  # pylint: disable=protected-access
        log.info("Attached CeleryCodeOwnerSpanProcessor")
    except ImportError:
        log.warning(
            "Unable to attach CeleryCodeOwnerSpanProcessor"
            " -- ddtrace module not found."
        )



"""
Publish events from a CSV to the Event Bus.

This is meant to help republish failed events. The CSV may be an export from Splunk, or it may be manually created, as
long as it has 'initial_topic', 'event_type', 'event_data_as_json', 'event_key_field', and 'event_metadata_as_json'
columns.

If the CSV is manually created, you will need to use the python __repr__ of all fields. For strings, including
json dumps, this means enclosing them in single quotes

Example row (second line split up for readability, in the csv it should be all one line)
initial_topic,event_type,event_data_as_json,event_key_field,event_metadata_as_json
'test-topic','org.openedx.test.event','{"user_data": {"id": 1, "is_active": True}}','user_data.id',
'{"event_type": "org.openedx.test.event", "id": "12345", "minorversion": 0, "source": "openedx/cms/web",
 "sourcehost": "ip-10-3-16-25", "time": "2023-08-10T17:15:38.549331+00:00", "sourcelib": [8, 5, 0]}'

This is created as a script instead of a management command because it's meant to be used as a one-off and not to
require pip installing this package into anything else to run. However, since edx-event-bus-kafka does expect certain
settings, the script must be run in an environment with DJANGO_SETTINGS_MODULE.

To run:
tox -e scripts -- python edx_arch_experiments/scripts/republish_failed_events.py
 --filename /Users/rgraber/oneoffs/failed_events.csv
"""

import csv
import json
import sys
from ast import literal_eval

import click
from openedx_events.event_bus import get_producer
from openedx_events.tooling import EventsMetadata, OpenEdxPublicSignal, load_all_signals


@click.command()
@click.option('--filename', type=click.Path(exists=True))
def read_and_send_events(filename):
    load_all_signals()
    producer = get_producer()
    try:
        log_columns = ['initial_topic', 'event_type', 'event_data_as_json', 'event_key_field', 'event_metadata_as_json']
        with open(filename) as log_file:
            reader = csv.DictReader(log_file)
            # Make sure csv contains all necessary columns for republishing
            missing_columns = set(log_columns).difference(set(reader.fieldnames))
            if len(missing_columns) > 0:
                print(f'Missing required columns {missing_columns}. Cannot republish events.')
                sys.exit(1)
            ids = set()
            for row in reader:
                # We log everything using __repr__, so strings get quotes around them and "None" gets
                # written literally. Use literal_eval to go from "None" to None and remove the extraneous quotes
                # from the logs
                empties = [key for key, value in row.items() if key in log_columns and literal_eval(value) is None]
                # If any row is missing data, stop processing the whole file to avoid sending events out of order
                if len(empties) > 0:
                    print(f'Missing required fields in row {reader.line_num}: {empties}. Will not continue publishing.')
                    sys.exit(1)

                topic = literal_eval(row['initial_topic'])
                event_type = literal_eval(row['event_type'])
                event_data = json.loads(literal_eval(row['event_data_as_json']))
                event_key_field = literal_eval(row['event_key_field'])
                metadata = EventsMetadata.from_json(literal_eval(row['event_metadata_as_json']))
                signal = OpenEdxPublicSignal.get_signal_by_type(event_type)
                if metadata.id in ids:
                    print(f"Skipping duplicate id {metadata.id}")
                    continue
                ids.add(metadata.id)

                producer.send(signal=signal, event_data=event_data, event_key_field=event_key_field, topic=topic,
                              event_metadata=metadata)
                print(f'Successfully published event to event bus. line={reader.line_num} {event_data=} {topic=}'
                      f' {event_key_field=} metadata={metadata.to_json()}')
    finally:
        producer.prepare_for_shutdown()


if __name__ == '__main__':
    read_and_send_events()



"""
Script to get the annotations from all failed checks in edx-platform after a given date

Gets all the commits to master after the date, then for each commit gets each check suite, then for each failed check
suite gets each run. Collects the annotations for all the failed runs. The annotations will sometimes contain useful
error messages, sometimes just the exit code. Getting the full logs requires admin permissions to edx-platform so it's
not included in this script.
Example output row:
commit_date,run_started_at,run_completed_at,commit_hash,name,message
2023-07-26T20:59:01Z,2023-07-27T06:56:23Z,2023-07-27T07:01:58Z,06e738e64a3485ecec037a9b8a36cf4ae145ea8a,
upgrade-one-python-dependency-workflow,Process completed with exit code 2.


This script takes a pretty long time to run (15m for 2 months) and there is a risk if you look too far back you will hit
your API limit.
"""

from csv import DictWriter
from datetime import datetime

import click
import requests


@click.command()
@click.option('--token', envvar='GITHUB_TOKEN')
@click.option('--start_date', type=click.DateTime(formats=["%Y-%m-%d"]), help="Date of earliest commit")
@click.option('--filename', help="Where to write the data")
def get_errors_from_date(token, start_date, filename):
    """
    Creates a csv documenting the annotations from all failed runs for commits to edx-platform after the given date

    Parameters:
        token (string): The GitHub API token. Retrieved from the env GITHUB_TOKEN variable
        start_date (date): The earliest date to look for
        filename (string): Where to write the csv

    """
    headers = {'Authorization': f"Bearer {token}"}
    all_commits_after_date = get_commits_after_date(start_date, headers=headers)
    all_check_suites = []
    all_rows = []
    for commit in all_commits_after_date:
        # gather all the check suite data from each commit into a single list
        add_commit_check_suites(commit, all_check_suites, headers)
    for check_suite in all_check_suites:
        # only record annotations for failed runs
        if check_suite['conclusion'] == 'failure':
            check_runs = requests.get(check_suite['check_runs_url'], headers=headers).json()
            for run in check_runs['check_runs']:
                if run['conclusion'] == 'failure' and run['output']['annotations_count'] > 0:
                    annotations = requests.get(run['output']['annotations_url'], headers=headers).json()
                    for annotation in annotations:
                        all_rows.append({
                            'commit_hash': run['head_sha'],
                            'name': run['name'],
                            'message': annotation['message'],
                            'run_started_at': run['started_at'],
                            'run_completed_at': run['completed_at'],
                            'commit_date': check_suite['commit_date']
                        })

    with open(filename, 'w') as f:
        writes = DictWriter(f, fieldnames=['commit_date', 'run_started_at', 'run_completed_at', 'commit_hash', 'name',
                                           'message'])
        writes.writeheader()
        writes.writerows(all_rows)


def get_commits_after_date(cut_off_date, headers):
    """
    Get API data for all commits to edx-platform/master after the given date

    Parameters:
        cut_off_date (date): Earliest date to look
        headers (dict): Authentication headers for GH requests

    Returns:
        A list of all the API responses for each commit after the date
    """
    base_url = "https://api.github.com/repos/openedx/edx-platform/commits?sha=master&per_page=100"
    # will keep track of whether we've hit our start_date. the API automatically returns commits ordered
    # by date, descending
    found_last = False
    all_commits_after_date = []
    page = 1
    while not found_last:
        page_url = f"{base_url}&page={page}"
        print(f"Fetching page {page_url}")
        response = requests.get(page_url, headers=headers)
        if response.status_code >= 400:
            print(response)
            break
        response_json = response.json()
        if len(response_json) == 0:
            break
        for single_commit in response_json:
            # if present, take off the "Z" at the end of the date to make it proper ISO format
            commit_date = datetime.fromisoformat(single_commit['commit']['committer']['date'].replace("Z", ""))
            if commit_date < cut_off_date:
                found_last = True
                break
            all_commits_after_date.append(single_commit)
        page += 1
    return all_commits_after_date


def add_commit_check_suites(current_commit, current_suites, headers):
    """
    Add API information from all check suites performed for a given commit to the given list

    Parameters:
        current_commit (str): the SHA of the commit to check
        current_suites (list): list to be extended
        headers (dict): Authentication headers for connecting to GitHub
    """
    sha = current_commit['sha']
    check_url = f"https://api.github.com/repos/openedx/edx-platform/commits/{sha}/check-suites?per_page=100"
    page = 1
    while True:
        # Keep going until we get an empty check_suites list or an error. An empty list means we've hit the last page.
        paginated_url = f"{check_url}&page={page}"
        print(f"Fetching page {paginated_url}")
        response = requests.get(paginated_url, headers=headers).json()
        if 'check_suites' not in response.keys():
            print(response)
            break
        check_suites = response['check_suites']
        if len(check_suites) == 0:
            break
        # silly line to pass the date of the commit along to eventually write in the spreadsheet
        current_suites.extend([{**s, 'commit_date': current_commit['commit']['committer']['date']}
                               for s in check_suites])
        page += 1


if __name__ == '__main__':
    get_errors_from_date()



"""
This script generates code owner mappings for monitoring LMS.

Sample usage::

    python lms/djangoapps/monitoring/scripts/generate_code_owner_mappings.py --repo-csv "Own Repos.csv"
     --app-csv "Own edx-platform Apps.csv" --dep-csv "Reference edx-platform Libs.csv" --override-csv "Overrides.csv"

Or for more details::

    python lms/djangoapps/monitoring/scripts/generate_code_owner_mappings.py --help


"""
import csv
import os
import re

import click

# Maps edx-platform installed Django apps to the edx repo that contains
# the app code. Please add in alphabetical order.
#
# The URLs here must match the URLs in the "Own: Repos" sheet:
# https://docs.google.com/spreadsheets/d/1qpWfbPYLSaE_deaumWSEZfz91CshWd3v3B7xhOk5M4U/view#gid=1990273504
# These applications are the ones which contain views we want to monitor
EDX_REPO_APPS = {
    'bulk_grades': 'https://github.com/openedx/edx-bulk-grades',
    'coaching': 'https://github.com/edx/platform-plugin-coaching',
    'completion': 'https://github.com/openedx/completion',
    'config_models': 'https://github.com/openedx/django-config-models',
    'consent': 'https://github.com/openedx/edx-enterprise',
    'csrf': 'https://github.com/openedx/edx-drf-extensions',
    'edx_arch_experiments': 'https://github.com/edx/edx-arch-experiments',
    'edx_name_affirmation': 'https://github.com/edx/edx-name-affirmation',
    'edx_proctoring': 'https://github.com/openedx/edx-proctoring',
    'edx_recommendations': 'https://github.com/edx/edx-recommendations',
    'edxval': 'https://github.com/openedx/edx-val',
    'enterprise': 'https://github.com/openedx/edx-enterprise',
    'enterprise_learner_portal': 'https://github.com/openedx/edx-enterprise',
    'eventtracking': 'https://github.com/openedx/event-tracking',
    'federated-content-connector': 'https://github.com/edx/federated-content-connector',
    'help_tokens': 'https://github.com/openedx/help-tokens',
    'integrated_channels': 'https://github.com/openedx/edx-enterprise',
    'learner_pathway_progress': 'https://github.com/edx/learner-pathway-progress',
    'learning_assistant': 'https://github.com/edx/learning-assistant',
    'lti_consumer': 'https://github.com/openedx/xblock-lti-consumer',
    'notices': 'https://github.com/edx/platform-plugin-notices',
    'openassessment': 'https://github.com/openedx/edx-ora2',
    'ora2': 'https://github.com/openedx/edx-ora2',
    'organizations': 'https://github.com/openedx/edx-organizations',
    'persona_integration': 'https://github.com/edx/persona-integration',
    'search': 'https://github.com/openedx/edx-search',
    'super_csv': 'https://github.com/openedx/super-csv',
    'translatable_xblocks': 'https://github.com/edx/translatable-xblocks',
    'wiki': 'https://github.com/openedx/django-wiki',
}

# Maps edx-platform installed Django apps to the third-party repo that contains
# the app code. Please add in alphabetical order.
#
# The URLs here must match the URLs in the "Reference: edx-platform Libs" sheet:
# https://docs.google.com/spreadsheets/d/1qpWfbPYLSaE_deaumWSEZfz91CshWd3v3B7xhOk5M4U/view#gid=506252353
# These applications are the ones which contain views we want to monitor
THIRD_PARTY_APPS = {
    'corsheaders': 'https://github.com/adamchainz/django-cors-headers',
    'django': 'https://github.com/django/django',
    'django_object_actions': 'https://github.com/crccheck/django-object-actions',
    'drf_yasg': 'https://github.com/axnsan12/drf-yasg',
    'edx_sga': 'https://github.com/mitodl/edx-sga',
    'lx_pathway_plugin': 'https://github.com/open-craft/lx-pathway-plugin',
    'oauth2_provider': 'https://github.com/jazzband/django-oauth-toolkit',
    'rest_framework': 'https://github.com/encode/django-rest-framework',
    'simple_history': 'https://github.com/treyhunner/django-simple-history',
    'social_django': 'https://github.com/python-social-auth/social-app-django',
}


@click.command()
@click.option(
    '--repo-csv',
    help="File name of .csv file with repo ownership details.",
    required=True
)
@click.option(
    '--app-csv',
    help="File name of .csv file with edx-platform app ownership details.",
    required=True
)
@click.option(
    '--dep-csv',
    help="File name of .csv file with edx-platform 3rd-party dependency ownership details.",
    required=True
)
@click.option(
    '--override-csv',
    help="File name of .csv file with ownership override details. Entries here will override entries in app-csv and"
         " dep-csv",
    required=False
)
def main(repo_csv, app_csv, dep_csv, override_csv=None):
    """
    Reads CSVs of ownership data and outputs config.yml setting to system.out.

    Expected Repo CSV format:

        \b
        repo url,owner.theme,owner.squad
        https://github.com/openedx/edx-bulk-grades,theme-a,team-red
        ...

    Expected App CSV format:

        \b
        Path,owner.theme,owner.squad
        ./openedx/core/djangoapps/user_authn,theme-b,team-blue
        ...

    Expected 3rd-party Dependency CSV format:

        \b
        repo url,owner.theme,owner.squad
        https://github.com/django/django,theme-a,team-red
        ...

    Expected Override CSV format:

        \b
        app_name,owner.theme,owner.squad
        integrated_channels,theme-b,team-blue
        ...

    Final output only includes paths which might contain views.

    """
    # Maps theme name to a list of code owners in the theme, and squad to full code owner name.
    # Code owner is a string combining theme and squad information.
    owner_map = {'theme_to_owners_map': {}, 'squad_to_theme_map': {}}
    # Maps owner names to a list of dotted module paths.
    # For example: { 'team-red': [ 'openedx.core.djangoapps.api_admin', 'openedx.core.djangoapps.auth_exchange' ] }
    owner_to_paths_map = {}
    _map_repo_apps('edx-repo', repo_csv, override_csv, EDX_REPO_APPS, owner_map, owner_to_paths_map)
    _map_repo_apps('3rd-party', dep_csv, override_csv, THIRD_PARTY_APPS, owner_map, owner_to_paths_map)
    _map_edx_platform_apps(app_csv, owner_map, owner_to_paths_map)

    # NB: An automated script looks for this comment when updating config files,
    # so please update regenerate_code_owners_config.py in jenkins-job-dsl-internal
    # if you change the comment format here.
    print(f'# Do not hand edit CODE_OWNER_MAPPINGS. Generated by {os.path.basename(__file__)}')
    print('CODE_OWNER_MAPPINGS:')
    for owner, path_list in sorted(owner_to_paths_map.items()):
        print(f"  {owner}:")
        path_list.sort()
        for path in path_list:
            print(f"  - {path}")

    owner_with_mappings_set = set(owner_to_paths_map.keys())
    print(f'# Do not hand edit CODE_OWNER_THEMES. Generated by {os.path.basename(__file__)}')
    print('CODE_OWNER_THEMES:')
    for theme, owner_list in sorted(owner_map['theme_to_owners_map'].items()):
        theme_owner_set = set(owner_list)
        # only include the theme's list of owners that have mappings
        theme_owner_with_mappings_list = list(theme_owner_set & owner_with_mappings_set)
        if theme_owner_with_mappings_list:
            print(f"  {theme}:")
            theme_owner_with_mappings_list.sort()
            for owner in theme_owner_with_mappings_list:
                print(f"  - {owner}")


def _map_repo_apps(csv_type, repo_csv, override_csv, app_to_repo_map, owner_map, owner_to_paths_map):
    """
    Reads CSV of repo ownership and uses app_to_repo_map to update owner_map and owner_to_paths_map

    Only the paths in app_to_repo_map will be added to owner_to_paths_map. All repositories not corresponding to an
    app in app_to_repo_map will be ignored.

    Arguments:
        csv_type (string): Either 'edx-repo' or '3rd-party' for error message
        repo_csv (string): File name for the edx-repo or 3rd-party repo csv
        override_csv (string): File name for the override csv (which may be None)
        app_to_repo_map (dict): Dict mapping Django apps to repo urls
        owner_map (dict): Dict of owner details
        owner_to_paths_map (dict): Holds results mapping owner to paths

    """
    with open(repo_csv) as file:
        csv_data = file.read()
    reader = csv.DictReader(csv_data.splitlines())

    overrides = {}
    if override_csv:
        with open(override_csv) as override_file:
            override_data = override_file.read()
        override_reader = csv.DictReader(override_data.splitlines())
        for row in override_reader:
            # transform "app, owner" rows into a dict of {app: owner}
            owner = _get_and_map_code_owner(row, owner_map)
            overrides[row['app']] = owner

    csv_repo_to_owner_map = {}
    for row in reader:
        owner = _get_and_map_code_owner(row, owner_map)
        csv_repo_to_owner_map[row.get('repo url')] = owner

    for app, repo_url in app_to_repo_map.items():
        # look for any overrides in the override map before looking in the spreadsheet
        owner = overrides.get(app, None)
        if not owner:
            owner = csv_repo_to_owner_map.get(repo_url, None)
        if owner:
            if owner not in owner_to_paths_map:
                owner_to_paths_map[owner] = []
            owner_to_paths_map[owner].append(app)
        else:
            raise Exception(
                f'ERROR: Repo {repo_url} was not found in {csv_type} csv. Needed for app {app}. '
                'Please reconcile the hardcoded lookup tables in this script with the ownership '
                'sheet.'
            )


def _map_edx_platform_apps(app_csv, owner_map, owner_to_paths_map):
    """
    Reads CSV of edx-platform app ownership and updates mappings
    """
    with open(app_csv) as file:
        csv_data = file.read()
    reader = csv.DictReader(csv_data.splitlines())
    for row in reader:
        path = row.get('Path')
        owner = _get_and_map_code_owner(row, owner_map)

        # add paths that may have views
        may_have_views = re.match(r'.*djangoapps', path) or re.match(r'[./]*openedx\/features', path)
        # remove cms (studio) paths and tests
        may_have_views = may_have_views and not re.match(r'.*(\/tests\b|cms\/).*', path)

        if may_have_views:
            path = path.replace('./', '')  # remove ./ from beginning of path
            path = path.replace('/', '.')  # convert path to dotted module name

            # skip catch-alls to ensure everything is properly mapped
            if path in ('common.djangoapps', 'lms.djangoapps', 'openedx.core.djangoapps', 'openedx.features'):
                continue

            if owner not in owner_to_paths_map:
                owner_to_paths_map[owner] = []
            owner_to_paths_map[owner].append(path)


def _get_and_map_code_owner(row, owner_map):
    """
    From a csv row, takes the theme and squad, update ownership maps, and return the code_owner.

    Will also warn if the squad appears in multiple themes.

    Arguments:
        row: A csv row that should have 'owner.theme' and 'owner.squad'.
        owner_map: A dict with 'theme_to_owners_map' and 'squad_to_theme_map' keys.

    Returns:
        The code_owner for the row.  This is made from the theme+squad (or squad if there is no theme).

    """
    theme = row.get('owner.theme')
    squad = row.get('owner.squad')
    assert squad, 'Csv row is missing required owner.squad: %s' % row

    # use lower case names only
    squad = squad.lower()
    if theme:
        theme = theme.lower()

    owner = f'{theme}-{squad}' if theme else squad
    theme = theme or squad

    if squad not in owner_map['squad_to_theme_map']:
        # store the theme for each squad for a later data integrity check
        owner_map['squad_to_theme_map'][squad] = theme

        # add to the list of owners for each theme
        if theme not in owner_map['theme_to_owners_map']:
            owner_map['theme_to_owners_map'][theme] = []
        owner_map['theme_to_owners_map'][theme].append(owner)

    # assert that squads have a unique theme. otherwise we have a data integrity issues in the csv.
    assert owner_map['squad_to_theme_map'][squad] == theme, \
        'Squad %s is associated with theme %s in row %s, but theme %s elsewhere in the csv.' % \
        (squad, theme, row, owner_map['squad_to_theme_map'][squad])

    return owner


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



"""
App for reporting configuration changes to Slack for operational awareness.
"""

from django.apps import AppConfig


class ConfigWatcher(AppConfig):
    """
    Django application to report configuration changes to operators.
    """
    name = 'edx_arch_experiments.config_watcher'

    # Mark this as a plugin app
    plugin_app = {}

    def ready(self):
        from .signals import receivers  # pylint: disable=import-outside-toplevel

        receivers.connect_receivers()



"""
Signal receivers for the config watcher.

Call ``connect_receivers`` to initialize.
"""

import html
import json
import logging
import urllib.request

from django.conf import settings
from django.db.models import signals
from django.dispatch import receiver
from django.utils.module_loading import import_string

log = logging.getLogger(__name__)

# .. setting_name: CONFIG_WATCHER_SLACK_WEBHOOK_URL
# .. setting_default: None
# .. setting_description: Slack webhook URL to send config change events to.
#   If not configured, this functionality is disabled.
CONFIG_WATCHER_SLACK_WEBHOOK_URL = getattr(settings, 'CONFIG_WATCHER_SLACK_WEBHOOK_URL', None)

# .. setting_name: CONFIG_WATCHER_SERVICE_NAME
# .. setting_default: None
# .. setting_description: Name of service, to be included in Slack messages in
#   in order to distinguish messages from multiple services being aggregated in
#   one channel. Can be a regular name ("LMS"), hostname, ("courses.example.com"),
#   or any other string. Optional.
CONFIG_WATCHER_SERVICE_NAME = getattr(settings, 'CONFIG_WATCHER_SERVICE_NAME', None)


def _send_to_slack(message):
    """Send this message as plain text to the configured Slack channel."""
    if not CONFIG_WATCHER_SLACK_WEBHOOK_URL:
        return

    if CONFIG_WATCHER_SERVICE_NAME:
        message = f"[{CONFIG_WATCHER_SERVICE_NAME}] {message}"

    # https://api.slack.com/reference/surfaces/formatting
    body_data = {
        'text': html.escape(message, quote=False)
    }

    req = urllib.request.Request(
        url=CONFIG_WATCHER_SLACK_WEBHOOK_URL, data=json.dumps(body_data).encode(),
        headers={"Content-Type": "application/json"},
    )
    with urllib.request.urlopen(req, timeout=2) as resp:
        status = resp.getcode()
        if status != 200:
            log.error(f"Slack rejected the config watcher message. {status=}, body={resp.read().decode('utf-8')}")


def _report_config_change(message):
    """
    Report this message string as a configuration change.

    Sends to logs and to Slack.
    """
    log.info(message)
    _send_to_slack(message)


def _report_waffle_change(model_short_name, instance, created, fields):
    """
    Report that a model instance has been created or updated.
    """
    verb = "created" if created else "updated"
    state_desc = ", ".join(f"{field}={repr(getattr(instance, field))}" for field in fields)
    _report_config_change(f"Waffle {model_short_name} {instance.name!r} was {verb}. New config: {state_desc}")


def _report_waffle_delete(model_short_name, instance):
    """
    Report that a model instance has been deleted.
    """
    _report_config_change(f"Waffle {model_short_name} {instance.name!r} was deleted")


# List of models to observe. Each is a dictionary that matches the
# keyword args of _register_waffle_observation.
_WAFFLE_MODELS_TO_OBSERVE = [
    {
        'model': 'waffle.models.Flag',
        'short_name': 'flag',
        'fields': ['everyone', 'percent', 'note'],
    },
    {
        'model': 'waffle.models.Switch',
        'short_name': 'switch',
        'fields': ['active', 'note'],
    },
    {
        'model': 'waffle.models.Sample',
        'short_name': 'sample',
        'fields': ['percent', 'note'],
    },
]


def _register_waffle_observation(*, model, short_name, fields):
    """
    Register a Waffle model for observation according to config values.

    Args:
        model (str): The model class to monitor, as a dotted string reference
        short_name (str): A short descriptive name for an instance of the model, e.g. "flag"
        fields (list): Names of fields to report on in the Slack message
    """

    model = import_string(model)

    # Note that weak=False is required here. Django by default only
    # holds weak references to receiver functions. But these inner
    # functions would then be garbage-collected, and Django would drop
    # them. So pass weak=False to make Django hold strong references
    # instead. (It works either way in devstack, apparently due to an
    # interaction with settings.DEBUG causing a reference to be held:
    # <https://stackoverflow.com/a/70651310>.)

    @receiver(signals.post_save, sender=model, weak=False, dispatch_uid=f"config_watcher_{short_name}_change")
    def report_waffle_change(*args, instance, created, **kwargs):
        try:
            _report_waffle_change(short_name, instance, created, fields)
        except:  # noqa pylint: disable=bare-except
            # Log and suppress error so Waffle change can proceed
            log.exception(f"Failed to report change to waffle {short_name}")

    @receiver(signals.post_delete, sender=model, weak=False, dispatch_uid=f"config_watcher_{short_name}_delete")
    def report_waffle_delete(*args, instance, **kwargs):
        try:
            _report_waffle_delete(short_name, instance)
        except:  # noqa pylint: disable=bare-except
            log.exception(f"Failed to report deletion of waffle {short_name}")

    log.info(f"Watching {model.__module__}.{model.__qualname__} for changes")


def connect_receivers():
    """
    Initialize application's receivers.
    """
    for config in _WAFFLE_MODELS_TO_OBSERVE:
        _register_waffle_observation(**config)



"""
Middleware for code_owner_2 custom attribute
"""
import logging

from django.urls import resolve
from edx_django_utils.monitoring import set_custom_attribute

from .utils import get_code_owner_from_module, is_code_owner_mappings_configured, set_code_owner_custom_attributes

log = logging.getLogger(__name__)


class CodeOwnerMonitoringMiddleware:
    """
    Django middleware object to set custom attributes for the owner of each view.

    For instructions on usage, see:
    https://github.com/edx/edx-arch-experiments/blob/master/edx_arch_experiments/datadog_monitoring/docs/how_tos/add_code_owner_custom_attribute_to_an_ida.rst

    Custom attributes set:
    - code_owner_2: The owning team mapped to the current view.
    - code_owner_2_module: The module found from the request or current transaction.
    - code_owner_2_path_error: The error mapping by path, if code_owner_2 isn't found in other ways.

    """
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)
        self._set_code_owner_attribute(request)
        return response

    def process_exception(self, request, exception):    # pylint: disable=W0613
        self._set_code_owner_attribute(request)

    def _set_code_owner_attribute(self, request):
        """
        Sets the code_owner_2 custom attribute for the request.
        """
        code_owner = None
        module = self._get_module_from_request(request)
        if module:
            code_owner = get_code_owner_from_module(module)

        if code_owner:
            set_code_owner_custom_attributes(code_owner)

    def _get_module_from_request(self, request):
        """
        Get the module from the request path or the current transaction.

        Side-effects:
            Sets code_owner_2_module custom attribute, used to determine code_owner_2.
            If module was not found, may set code_owner_2_path_error custom attribute
                if applicable.

        Returns:
            str: module name or None if not found

        """
        if not is_code_owner_mappings_configured():
            return None

        module, path_error = self._get_module_from_request_path(request)
        if module:
            set_custom_attribute('code_owner_2_module', module)
            return module

        # monitor errors if module was not found
        if path_error:
            set_custom_attribute('code_owner_2_path_error', path_error)
        return None

    def _get_module_from_request_path(self, request):
        """
        Uses the request path to get the view_func module.

        Returns:
            (str, str): (module, error_message), where at least one of these should be None

        """
        try:
            view_func, _, _ = resolve(request.path)
            module = view_func.__module__
            return module, None
        except Exception as e:  # pragma: no cover, pylint: disable=broad-exception-caught
            return None, str(e)



"""
Tests for the code_owner monitoring middleware
"""
import timeit
from unittest import TestCase
from unittest.mock import call, patch

import ddt
from django.test import override_settings

from edx_arch_experiments.datadog_monitoring.code_owner.utils import (
    clear_cached_mappings,
    get_code_owner_from_module,
    set_code_owner_attribute_from_module,
)


@ddt.ddt
class MonitoringUtilsTests(TestCase):
    """
    Tests for the code_owner monitoring utility functions
    """
    def setUp(self):
        super().setUp()
        clear_cached_mappings()

    @override_settings(CODE_OWNER_MAPPINGS={
        'team-red': [
            'openedx.core.djangoapps.xblock',
            'lms.djangoapps.grades',
        ],
        'team-blue': [
            'common.djangoapps.xblock_django',
        ],
    })
    @ddt.data(
        ('xbl', None),
        ('xblock_2', None),
        ('xblock', 'team-red'),
        ('openedx.core.djangoapps', None),
        ('openedx.core.djangoapps.xblock', 'team-red'),
        ('openedx.core.djangoapps.xblock.views', 'team-red'),
        ('grades', 'team-red'),
        ('lms.djangoapps.grades', 'team-red'),
        ('xblock_django', 'team-blue'),
        ('common.djangoapps.xblock_django', 'team-blue'),
    )
    @ddt.unpack
    def test_code_owner_mapping_hits_and_misses(self, module, expected_owner):
        actual_owner = get_code_owner_from_module(module)
        self.assertEqual(expected_owner, actual_owner)

    @override_settings(CODE_OWNER_MAPPINGS=['invalid_setting_as_list'])
    @patch('edx_arch_experiments.datadog_monitoring.code_owner.utils.log')
    def test_code_owner_mapping_with_invalid_dict(self, mock_logger):
        with self.assertRaises(TypeError):
            get_code_owner_from_module('xblock')

        mock_logger.exception.assert_called_with(
            'Error processing CODE_OWNER_MAPPINGS. list indices must be integers or slices, not str',
        )

    def test_code_owner_mapping_with_no_settings(self):
        self.assertIsNone(get_code_owner_from_module('xblock'))

    def test_code_owner_mapping_with_no_module(self):
        self.assertIsNone(get_code_owner_from_module(None))

    def test_mapping_performance(self):
        code_owner_mappings = {
            'team-red': []
        }
        # create a long list of mappings that are nearly identical
        for n in range(1, 200):
            path = f'openedx.core.djangoapps.{n}'
            code_owner_mappings['team-red'].append(path)
        with override_settings(CODE_OWNER_MAPPINGS=code_owner_mappings):
            call_iterations = 100
            time = timeit.timeit(
                # test a module name that matches nearly to the end, but doesn't actually match
                lambda: get_code_owner_from_module('openedx.core.djangoapps.XXX.views'), number=call_iterations
            )
            average_time = time / call_iterations
            self.assertLess(average_time, 0.0005, f'Mapping takes {average_time}s which is too slow.')

    @override_settings(CODE_OWNER_MAPPINGS={
        'team-red': ['edx_arch_experiments.datadog_monitoring.tests.code_owner.test_utils']
    })
    @patch('edx_arch_experiments.datadog_monitoring.code_owner.utils.set_custom_attribute')
    def test_set_code_owner_attribute_from_module_success(self, mock_set_custom_attribute):
        set_code_owner_attribute_from_module(__name__)
        self._assert_set_custom_attribute(mock_set_custom_attribute, code_owner='team-red', module=__name__)

    def _assert_set_custom_attribute(self, mock_set_custom_attribute, code_owner, module, check_theme_and_squad=False):
        """
        Helper to assert that the proper set_custom_metric calls were made.
        """
        call_list = []
        if code_owner:
            call_list.append(call('code_owner_2', code_owner))
            if check_theme_and_squad:
                call_list.append(call('code_owner_2_theme', code_owner.split('-')[0]))
                call_list.append(call('code_owner_2_squad', code_owner.split('-')[1]))
        call_list.append(call('code_owner_2_module', module))
        mock_set_custom_attribute.assert_has_calls(call_list, any_order=True)



"""
App for running answer submissions inside of codejail.
"""

from django.apps import AppConfig
from edx_django_utils.plugins.constants import PluginURLs


class CodejailService(AppConfig):
    """
    Django application to run things in codejail.
    """
    name = 'edx_arch_experiments.codejail_service'

    plugin_app = {
        PluginURLs.CONFIG: {
            'lms.djangoapp': {
                PluginURLs.NAMESPACE: 'codejail_service',
            }
        },
    }



"""
Tests for datadog span processor.
"""
from unittest.mock import patch

from django.test import TestCase

from edx_arch_experiments.datadog_monitoring.code_owner.datadog import CeleryCodeOwnerSpanProcessor


class FakeSpan:
    """
    A fake Span instance with span name and resource.
    """

    def __init__(self, name, resource):
        self.name = name
        self.resource = resource


class TestCeleryCodeOwnerSpanProcessor(TestCase):
    """
    Tests for CeleryCodeOwnerSpanProcessor.
    """

    @patch('edx_arch_experiments.datadog_monitoring.code_owner.utils.set_custom_attribute')
    def test_celery_span(self, mock_set_custom_attribute):
        """ Tests processor with a celery span. """
        proc = CeleryCodeOwnerSpanProcessor()
        celery_span = FakeSpan('celery.run', 'test.module.for.celery.task')

        proc.on_span_start(celery_span)

        mock_set_custom_attribute.assert_called_once_with('code_owner_2_module', 'test.module.for.celery.task')

    @patch('edx_arch_experiments.datadog_monitoring.code_owner.utils.set_custom_attribute')
    def test_other_span(self, mock_set_custom_attribute):
        """ Tests processor with a non-celery span. """
        proc = CeleryCodeOwnerSpanProcessor()
        celery_span = FakeSpan('other.span', 'test.resource.name')

        proc.on_span_start(celery_span)

        mock_set_custom_attribute.assert_not_called()

    @patch('edx_arch_experiments.datadog_monitoring.code_owner.utils.set_custom_attribute')
    def test_non_span(self, mock_set_custom_attribute):
        """ Tests processor with an object that doesn't have span name or resource. """
        proc = CeleryCodeOwnerSpanProcessor()
        non_span = object()

        proc.on_span_start(non_span)

        mock_set_custom_attribute.assert_not_called()



"""
Codejail service API.
"""

import json
import logging
from copy import deepcopy

from codejail.safe_exec import SafeExecException, safe_exec
from django.conf import settings
from edx_toggles.toggles import SettingToggle
from jsonschema.exceptions import best_match as json_error_best_match
from jsonschema.validators import Draft202012Validator
from rest_framework.decorators import api_view, parser_classes, permission_classes
from rest_framework.parsers import FormParser, MultiPartParser
from rest_framework.permissions import IsAdminUser
from rest_framework.response import Response

log = logging.getLogger(__name__)

# .. toggle_name: CODEJAIL_SERVICE_ENABLED
# .. toggle_implementation: SettingToggle
# .. toggle_default: True
# .. toggle_description: If True, codejail execution calls will be accepted over the network,
#   allowing this IDA to act as a codejail service for another IDA.
# .. toggle_use_cases: circuit_breaker
# .. toggle_creation_date: 2023-12-21
# .. toggle_tickets: https://github.com/openedx/edx-platform/issues/33538
CODEJAIL_SERVICE_ENABLED = SettingToggle('CODEJAIL_SERVICE_ENABLED', default=True, module_name=__name__)

# Schema for the JSON passed in the v0 API's 'payload' field.
payload_schema = {
    'type': 'object',
    'properties': {
        'code': {'type': 'string'},
        'globals_dict': {'type': 'object'},
        # Some of these are configured as union types because
        # edx-platform appears to currently default to None for some
        # of them (rather than omitting the keys.)
        'python_path': {
            'anyOf': [
                {
                    'type': 'array',
                    'items': {'type': 'string'},
                },
                {'type': 'null'},
            ],
        },
        'limit_overrides_context': {
            'anyOf': [
                {'type': 'string'},
                {'type': 'null'},
            ],
        },
        'slug': {
            'anyOf': [
                {'type': 'string'},
                {'type': 'null'},
            ],
        },
        'unsafely': {'type': 'boolean'},
    },
    'required': ['code', 'globals_dict'],
}
# Use this rather than jsonschema.validate, since that would check the schema
# every time it is called. Best to do it just once at startup.
Draft202012Validator.check_schema(payload_schema)
payload_validator = Draft202012Validator(payload_schema)


# A note on the authorization model used here:
#
# We really just need one service account to be able to call this, and
# then also allow is_staff to call it for convenience and debugging
# purposes.
#
# To do this "right", I'd probably have to create an empty abstract
# model, create a permission on it, create a group, add the permission
# to the group, and add the service account to the group. Then I could
# check the calling user's has_perm. If I wanted to use bridgekeeper
# (as we're trying to do more of) I might be able to give bridgekeeper
# a `@blanket_rule` that checks membership in the group, then use
# bridgekeeper here instead of checking permissions directly, but it's
# possible this wouldn't work because bridgekeeper might require there
# to be a model instance to pass in (and there wouldn't be one, since
# it's just an abstract model.)
#
# But... given that the service account will be is_staff, and I'll be
# opening this up to is_staff alongside the intended service account,
# and this is already a hacky intermediate solution... we can just use
# the `IsAdminUser` permission class and be done with it.


@api_view(['POST'])
@parser_classes([FormParser, MultiPartParser])
@permission_classes([IsAdminUser])
def code_exec_view_v0(request):
    """
    Executes code in a codejail sandbox for a remote caller.

    This implements the API used by edxapp's xmodule.capa.safe_exec.remote_exec.
    It accepts a POST of a form containing a `payload` value and zero or more
    extra files.

    The payload is JSON and contains the parameters to be sent to codejail's
    safe_exec (aside from `extra_files`). See payload_schema for type information.

    This API does not permit `unsafely=true`.

    If the response is a 200, the codejail execution completed. The response
    will be JSON containing either a single key `globals_dict` (containing
    the global scope values at the end of a run to completion) or `emsg` (the
    exception the submitted code raised).

    Other responses are errors, with a JSON body containing further details.
    """
    if not CODEJAIL_SERVICE_ENABLED.is_enabled():
        return Response({'error': "Codejail service not enabled"}, status=500)

    # There's a risk of getting into a loop if e.g. the CMS asks the
    # LMS to run codejail executions on its behalf, and the LMS is
    # *also* inadvertently configured to call the LMS (itself).
    # There's no good reason to have a chain of >2 services passing
    # codejail requests along, so only allow execution here if we
    # aren't going to pass it along to someone else.
    if getattr(settings, 'ENABLE_CODEJAIL_REST_SERVICE', False):
        log.error(
            "Refusing to run codejail request from over the network "
            "when we're going to pass it to another IDA anyway"
        )
        return Response({'error': "Codejail service is misconfigured. (Refusing to act as relay.)"}, status=500)

    params_json = request.data['payload']
    params = json.loads(params_json)

    if json_error := json_error_best_match(payload_validator.iter_errors(params)):
        return Response({'error': f"Payload JSON did not match schema: {json_error.message}"}, status=400)

    complete_code = params['code']  # includes standard prolog
    input_globals_dict = params['globals_dict']
    python_path = params.get('python_path') or []
    limit_overrides_context = params.get('limit_overrides_context')
    slug = params.get('slug')
    unsafely = params.get('unsafely')

    # Convert to a list of (string, bytestring) pairs. Any duplicated file names
    # are resolved as last-wins.
    extra_files = [(filename, file.read()) for filename, file in request.FILES.items()]

    # Far too dangerous to allow unsafe executions to come in over the
    # network, no matter who we think the caller is. The caller is the
    # one who has the context on safety.
    if unsafely:
        return Response({'error': "Refusing codejail execution with unsafely=true"}, status=400)

    output_globals_dict = deepcopy(input_globals_dict)  # Output dict will be mutated by safe_exec
    try:
        safe_exec(
            complete_code,
            output_globals_dict,
            python_path=python_path,
            extra_files=extra_files,
            limit_overrides_context=limit_overrides_context,
            slug=slug,
        )
    except SafeExecException as e:
        log.debug("CodejailService execution failed for {slug=} with: {e}")
        return Response({'emsg': str(e)})

    log.debug("CodejailService execution succeeded for {slug=}, with globals={output_globals_dict!r}")
    return Response({'globals_dict': output_globals_dict})



"""
App for 2U-specific edx-platform Datadog monitoring.
"""
from django.apps import AppConfig


class DatadogMonitoring(AppConfig):
    """
    Django application to handle 2U-specific Datadog monitoring.
    """
    name = 'edx_arch_experiments.datadog_monitoring'

    # Mark this as a plugin app
    plugin_app = {}

    def ready(self):
        # Implicitly connect signal handlers decorated with @receiver
        # pylint: disable=import-outside-toplevel,unused-import
        from edx_arch_experiments.datadog_monitoring.signals import handlers
